(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`AMPS`", {
(*"PhysicalConstants`","Units`",*)
    "AtomicDensityMatrix`",
    "AtomicDensityMatrix`Common`",
    "AtomicDensityMatrix`CouplingCoefficients`",
    "AtomicDensityMatrix`TensorAlgebra`",
    "AtomicDensityMatrix`AtomicSystem`",
    "AtomicDensityMatrix`ExperimentalSystem`",
    "AtomicDensityMatrix`DensityMatrix`",
    "AtomicDensityMatrix`PolarizationMoments`",
    "AtomicDensityMatrix`WignerEckart`",
    "AtomicDensityMatrix`WignerRotate`"}];

Begin["`Private`"];

(*Angular momentum probability distribution*)

(*Angular momentum probability distribution in terms of the \[Kappa]-q tensor elements [Ale2004, Eq. (A11)]*)

\[Rho]JJ = MaxProjectionProbability;

MaxProjectionProbability[T_?TensorDecompositionQ, th_, ph_] := Module[
    {J = Rank[T], T1 = ToContravariant@T},
    Sqrt[(4 Pi) / (2J + 1)] Sum[
        CG[{J, J}, {k, 0}, {J, J}] Component[T1, k, q] SphericalHarmonicY[k, q, th, ph],
        {k, 0, 2J}, {q, -k, k}
    ]
];

(*Angular momentum probability distribution in terms of the m m' tensor elements [Ale2004, Eq. (A12)]*)

MaxProjectionProbability[rho_?OperatorQ, th_, ph_] :=
    (ECT[#]\[CenterDot]rho\[CenterDot]#&)[WignerD[Rank[rho], {ph, th, 0}]\[CenterDot]StretchedState[Rank[rho]]];

Options[AMPSPlot] = Options[SphericalPlot3D];

SetOptions[AMPSPlot, Boxed -> False, Axes -> False, PlotRange -> All];

AMPSPlot[rho_, opts : OptionsPattern[]] :=
    SphericalPlot3D[Evaluate@Re@MaxProjectionProbability[rho, th, ph], {th, 0, Pi}, {ph, 0, 2Pi}, opts, Evaluate@Options@AMPSPlot];

End[];

EndPackage[];
