(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`PolarizationMoments`", {
(*"PhysicalConstants`","Units`",*)
    "AtomicDensityMatrix`",
    "AtomicDensityMatrix`Common`",
    "AtomicDensityMatrix`CouplingCoefficients`",
    "AtomicDensityMatrix`TensorAlgebra`",
    "AtomicDensityMatrix`AtomicSystem`",
    "AtomicDensityMatrix`ExperimentalSystem`",
    "AtomicDensityMatrix`DensityMatrix`",
    "AtomicDensityMatrix`WignerEckart`",
    "AtomicDensityMatrix`WignerRotate`"
}];

Begin["`Private`"];

Options[PolarizationMoments] = DeleteCases[Options[DensityMatrix] /. (r_ -> _) -> (r -> Automatic), Representation -> _];

PolarizationMoments[j_?HalfIntegerQ, opts : OptionsPattern[]] := PolarizationMoments[ZeemanSystem[j], opts];

PolarizationMoments[label_, j_?HalfIntegerQ, opts : OptionsPattern[]] := PolarizationMoments[ZeemanSystem[label, j], opts];

PolarizationMoments[sys : {_AtomicState..}, opts : OptionsPattern[]] :=
    Module[{newopts, rfunc, sym, lab, tv, td},
        newopts = removeAutomaticOptions[PolarizationMoments, opts];
        {sym, lab, tv, td} = OptionValue[DensityMatrix, newopts, {DMSymbol, DMLabel, TimeVariable, TimeDependence}];
        rfunc[l1_, l2_, k_, q_] =
            (Subscript[sym, lab /. None | False -> Sequence[], {l1, l2}, {k, q}][tv])[[If[td, All, 0]]];
        DirectSelfProduct[
            Table[
                Transpose@{Table[rfunc[#1[[1]], #2[[1]], k, q], {q, k, -k, -1}]},
                {k, Abs[#1[[2]] - #2[[2]]], #1[[2]] + #2[[2]]}
            ]&,
            {LevelLabel[#], F[#]}& /@ RemoveZeemanSublevels[sys]
        ]
    ];

PolarizationMatrices[j_?HalfIntegerQ] := Component@PolarizationMatrices[ZeemanSystem[j]];

PolarizationMatrices[sys : {_AtomicState ..}] := Module[{r},
    ECT[PolarizationMoments[sys, TimeDependence -> False, DMSymbol -> r]] /.
        {{___, Subscript[r, {s1_, s2_}, {k_, 0}], ___}} :> WignerEckart[sys, {Polarization[s1, s2], k}]
];

(*Convert between m m' and \[Kappa] q representation (contravariant components) using the polarization operator. [Var88, Sec. 6.1.5, Eqs. (49,50)], also [Ale2004, Eqs. (A3-8) ]*)

Decomposition[sys : {_AtomicState..}, x_?OperatorQ] :=
    Map[Tr[#\[CenterDot]x]&, ECT@PolarizationMatrices[sys], {5}];

Decomposition[x_?OperatorQ] := Component@Decomposition[ZeemanSystem@Rank@x, x];

Recomposition[x_?TensorDecompositionQ] := Recomposition[ZeemanSystem@Rank@x, {{x}}];

Recomposition[j_?HalfIntegerQ, x_?TensorDecompositionQ] := Recomposition[ZeemanSystem@j, {{x}}];

Recomposition[sys : {_AtomicState..}, x_?ContravariantDecompositionMatrixQ] := PolarizationMatrices[sys]\[CenterDot]x;

Recomposition[sys : {_AtomicState..}, x_?CovariantDecompositionMatrixQ] := x\[CenterDot]ECT@PolarizationMatrices[sys];

Recomposition[j_?HalfIntegerQ, x_?ContravariantQ] := Component[PolarizationMatrices[j], Rank[x]]\[CenterDot]x;

Recomposition[j_?HalfIntegerQ, x_?CovariantQ] := x\[CenterDot]ECT@Component[PolarizationMatrices[j], Rank[x]];

End[];

EndPackage[];
