(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`TensorAlgebra`",
    { "AtomicDensityMatrix`",
        "AtomicDensityMatrix`Common`",
        "AtomicDensityMatrix`CouplingCoefficients`",
        "AtomicDensityMatrix`AtomicSystem`"}];

(* Exported symbols added here with SymbolName::usage *)

StretchedState::usage = "";

(*Testing and component extraction*)

ContravariantQ::usage = "ContravariantQ[expr] gives True if expr is in the form of a contravariant spherical tensor or state vector, i.e., a n x 1 array (the elements of which could be operators), and gives False otherwise.";

CovariantQ::usage = "CovariantQ[expr] gives True if expr is in the form of a covariant spherical tensor or state vector, i.e., a 1 x n array (the elements of which could be operators), and gives False otherwise.";

SphericalTensorQ::usage = "SphericalTensorQ[expr] gives True if expr is in the form of a covariant or contravariant spherical tensor or state vector, i.e., a 1 x n or n x 1 array (the elements of which could be operators), and gives False otherwise.";

TensorDecompositionQ::usage = "TensorDecompositionQ[expr] gives True if expr is in the form of a list of spherical tensors of rank m,m+1,...,n, for some m>n, and gives False otherwise.";

CovariantDecompositionQ::usage = "";

ContravariantDecompositionQ::usage = "";

TensorDecompositionMatrixQ::usage = "TensorDecompositionMatrixQ[expr] gives True if expr is a matrix of tensor decompositions, and gives False otherwise.";

CovariantDecompositionMatrixQ::usage = "";

ContravariantDecompositionMatrixQ::usage = "";

OperatorQ::usage = "OperatorQ[expr] gives True if expr is in the form of a matrix operator, i.e., an n x n matrix, and gives False otherwise.";

ScalarQ::usage = "ScalarQ[expr] gives True if expr is not in the form of a Cartesian vector, spherical tensor, matrix operator or tensor decomposition, and gives False otherwise.";

CartesianQ::usage = "CartesianQ[expr] gives True if expr is in the form of a Cartesian vector, i.e., a one-dimensional list (the elements of which could be operators), and gives False otherwise.";

OneByOneQ::usage = "";

Rank::usage = "";

Component::usage = "";

TensorDecompositionComponent::usage = "";

SphericalTensorComponent::usage = "";

Components::usage = "";

MapOnComponents::usage = "";

(*Display*)

TensorForm::usage = "";

(*Matrix multiplication*)

MatrixMultiply::usage = "MatrixMultiply[a,b,c] gives inner and outer products of covariant and contravariant spherical tensors, bras and kets, operators, tensor decompositions of operators, Cartesian vectors, and scalars.";

TensorNorm::usage = "TensorNorm[a] gives the norm of a spherical tensor or Cartesian vector.";

ContractOneByOnes::usage = "";

SumRanks::usage = "";

(*Conjugation*)

ExplicitConjugate::usage = "ExplicitConjugate[expr] takes the complex conjugate of expr, explicitly assuming that all variables are real.";

EC::usage = "EC[expr] takes the complex conjugate of expr, explicitly assuming that all variables are real.";

ExplicitConjugateTranspose::usage = "ExplicitConjugateTranspose[expr] takes the Hermitian conjugate of a variety of types of expressions, explicitly assuming that all variables are real. If expr is a spherical tensor or matrix operator, the conjugate transpose is taken (converting covariant to contravariant, and vice versa). If expr is a tensor decomposition, the conjugate transpose of each tensor in the decomposition is taken. If expr is a Cartesian vector or a scalar, the complex conjugate is taken.";

ECT::usage = "ECT[expr] takes the Hermitian conjugate of a variety of types of expressions, explicitly assuming that all variables are real. If expr is a spherical tensor or matrix operator, the conjugate transpose is taken (converting covariant to contravariant, and vice versa). If expr is a tensor decomposition, the conjugate transpose of each tensor in the decomposition is taken. If expr is a Cartesian vector or a scalar, the complex conjugate is taken.";

GeneralizedConjugateTranspose::usage = "GeneralizedConjugateTranspose[expr] takes the Hermitian conjugate of a variety of types of expressions. If expr is a spherical tensor or matrix operator, the conjugate transpose is taken (converting covariant to contravariant, and vice versa). If expr is a tensor decomposition, the conjugate transpose of each tensor in the decomposition is taken. If expr is a Cartesian vector or a scalar, the complex conjugate is taken.";

GCT::usage = "";

ToCovariant::usage = "ToCovariant[expr] converts any spherical tensor or Cartesian vector to a covariant spherical tensor.";

ToContravariant::usage = "ToContravariant[expr] converts any spherical tensor or Cartesian vector to a contravariant spherical tensor.";

(*Dual*)

DualMatrix::usage = "";

Dual::usage = "";

TensorConjugate::usage = "";

(*Spherical to Cartesian conversion*)

ToCartesian::usage = "ToCartesian[expr] converts expr from a rank-one spherical tensor to a Cartesian vector. If expr is already a Cartesian vector it is left unchanged.";

ToCartesianMatrix::usage = "";

(*Tensor product*)

SphericalTensorProduct::usage = "";

(*Error messages*)

MatrixMultiply::"coco" = "Inner product attempted between two covariant or two contravariant tensors.";

MatrixMultiply::"opco" = "Operator multiplication attempted on wrong side of covariant or contravariant tensor.";

MatrixMultiply::"sphca" = "Inner product attempted between spherical and Cartesian tensors.";

Begin["`Private`"];

StretchedState[j_] := ECT@{Table[KroneckerDelta[1, i], {i, 2j + 1}]};

(*Testing and component extraction*)

(*Check for various structures*)

(*A covariant spherical tensor is a 1xn (horizontal) array, where the elements are either all not lists, or all square matrices (operators) *)
CovariantQ[expr_] := ArrayQ[expr] && MatchQ[Dimensions[expr], {1, _} | {1, _, x_, x_}];

CovariantQ[expr_, test_] := MatchQ[expr, {{__?test}}];

ContravariantQ[expr_] :=
    ArrayQ[expr] && MatchQ[Dimensions[expr], {_, 1} | {_, 1, x_, x_}];

ContravariantQ[expr_, test_] := MatchQ[expr, {{_?test}..}];

SphericalTensorQ[expr_] := CovariantQ[expr] || ContravariantQ[expr];

SphericalTensorQ[expr_, test_] :=
    CovariantQ[expr, test] || ContravariantQ[expr, test];

CartesianQ[expr_] := ArrayQ[expr, 1 | 3];

CartesianQ[expr_, test_] := ArrayQ[expr, 1 | 3] && MatchQ[expr, {__?test}];

OperatorQ[expr_] := MatrixQ@expr && Equal @@ Dimensions@expr;

decompositionQ[expr_, test_] :=
    Head[expr] === List && (And @@ test /@ expr) &&
        (# - First@# &)[SphericalTensorRank /@ expr] === Range@Length@expr - 1;

CovariantDecompositionQ[expr_] := decompositionQ[expr, CovariantQ];

ContravariantDecompositionQ[expr_] := decompositionQ[expr, ContravariantQ];

TensorDecompositionQ[expr_] := decompositionQ[expr, SphericalTensorQ];

TensorDecompositionMatrixQ[expr_] := MatrixQ[expr, TensorDecompositionQ];

CovariantDecompositionMatrixQ[expr_] := MatrixQ[expr, CovariantDecompositionQ];

ContravariantDecompositionMatrixQ[expr_] := MatrixQ[expr, ContravariantDecompositionQ];

OneByOneQ[expr_] := MatchQ[expr, {{_}}];

ScalarQ[expr_] := (Head[expr] =!= List && Head[expr] =!= Operator);

(*Returns the rank \[Kappa] of an irreducible tensor (a^\[Kappa]). For a reducible tensor a in the m m' or \[Kappa] q basis, returns the value J, where 2J is the rank of the highest irreducible tensor moment possible in the irreducible tensor decomposition of a. If a is written as a direct product of two irreducible tensors, a=A^J B^J, the irreducible tensors have rank J.*)

Rank[a_?SphericalTensorQ] := SphericalTensorRank[a];

Rank[a_?CartesianQ] := CartesianRank[a];

Rank[a_?OperatorQ] := OperatorRank[a];

Rank[a_?TensorDecompositionQ] := TensorDecompositionRank[a];

SphericalTensorRank[a_] := (Length@SphericalTensorComponents[a] - 1) / 2;

CartesianRank[a_] := (Length@CartesianComponents[a] - 1) / 2;

OperatorRank[a_] := (Length[a] - 1) / 2;

TensorDecompositionRank[a_] := (Length[a] - 1) / 2;

(*Extract or operate on all components of a vector or tensor*)

MapOnComponents[f_, expr_?CartesianQ] := f /@ expr;

MapOnComponents[f_, expr_?SphericalTensorQ] := Map[f, expr, {2}];

MapOnComponents[f_, expr_?TensorDecompositionQ] := f /@ expr;

MapOnComponents[f_, expr_?OperatorQ] := Map[f, expr, {2}];

MapOnComponents[f_, expr_?TensorDecompositionMatrixQ] := Map[f, expr, {2}];

MapOnComponents[f_, expr_?ScalarQ] := f[expr];

Components[a_?SphericalTensorQ] := SphericalTensorComponents[a];

Components[a_?CartesianQ] := CartesianComponents[a];

Components[a_?OperatorQ] := OperatorComponents[a];

SphericalTensorComponents[a_] := Flatten[a, 1];

CartesianComponents[a_] := a;

OperatorComponents[a_] := Flatten[a, 2];

Components[system : {_AtomicState..}, op_?OperatorQ, crit__Rule] :=
    Map[Extract[op, #]&, DirectSelfProduct[{#1, #2}&, Flatten[StatePosition[system, crit]]], {2}];

(*Individual components of vectors, tensors, or operators.*)

Component[a_?OneByOneQ] := a[[1, 1]];

Component[a_?TensorDecompositionQ, k_Integer] :=
    TensorDecompositionComponent[a, k];

Component[a_?TensorDecompositionQ, k_Integer, q_Integer] :=
    SphericalTensorComponent[TensorDecompositionComponent[a, k], q];

TensorDecompositionComponent[a_, k_Integer, q_Integer] :=
    SphericalTensorComponent[TensorDecompositionComponent[a, k], q];

TensorDecompositionComponent[a_, k_Integer] :=
    Select[a, SphericalTensorRank[#] == k&, 1][[1]];

Component[a_ /; OperatorQ[a] || OneByOneQ[a], m1_, m2_] :=
    a[[-(m1 + OperatorRank[a] + 1), -(m2 + OperatorRank[a] + 1)]];

Component[a_?SphericalTensorQ, q_ /; IntegerQ[2q]] :=
    SphericalTensorComponent[a, q];

SphericalTensorComponent[a_, q_ /; IntegerQ[2q]] :=
    SphericalTensorComponents[a][[-(q + SphericalTensorRank[a] + 1)]];

Component[a_?CartesianQ, i_?IntegerQ] := a[[i]];

Component[system : {_AtomicState..}, op_?OperatorQ, crit__Rule] :=
    Components[system, op, crit][[1, 1]];

Component[system : {_AtomicState..}, op_?OperatorQ, {crit1__Rule}, {crit2__Rule}] :=
    Extract[op, {First[Flatten[StatePosition[system, crit1]]], First[Flatten[StatePosition[system, crit2]]]}];

Component[sys : {_AtomicState..}, op_?OperatorQ, {s1_AtomicState, s2_AtomicState}] :=
    op[[Position[sys, s1][[1, 1]], Position[sys, s2][[1, 1]]]];

Component[sys : {_AtomicState..}, op_?OperatorQ, {s1_AtomicState, s2_AtomicState}, kq__?IntegerQ] :=
    Component[Component[sys, op, {s1, s2}], kq];

Component[sys : {_AtomicState..}, op_, crit1_Rule, crit2_Rule, kq__Integer] :=
    Component[Component[sys, op, {crit1}, {crit2}], kq];

Component[sys : {_AtomicState..}, op_, {crit1__Rule}, {crit2__Rule}, k_Integer] :=
    Component[Component[sys, op, {crit1}, {crit2}], k];

Component[sys : {_AtomicState..}, op_, {crit1__Rule}, {crit2__Rule}, k_Integer, q_Integer] :=
    Component[Component[sys, op, {crit1}, {crit2}], k, q];

Component[sys : {_AtomicState..}, op_, crit1_Rule, crit2_Rule] /; TensorDecompositionMatrixQ[op] :=
    TensorDecompositionMatrixComponent[sys, op, {crit1}, {crit2}];

Component[sys : {_AtomicState..}, op_?TensorDecompositionMatrixQ, {crit1__Rule}, {crit2__Rule}] :=
    TensorDecompositionMatrixComponent[sys, op, {crit1}, {crit2}];

TensorDecompositionMatrixComponent[sys : {_AtomicState..}, op_, {crit1__Rule}, {crit2__Rule}] :=
    Module[{sys1 = RemoveZeemanSublevels@sys},
        op[[StatePosition[sys1, crit1][[1, 1]],
            StatePosition[sys1, crit2][[1, 1]]]]
    ];

(*ReplaceMatrixElement[system : {_AtomicState..}, op_?OperatorQ, rep : _Rule | {__Rule}, {crit1_, crit2_}] :=*)
    (*Module[{pos1 = Flatten@StatePosition[system, crit1],*)
        (*pos2 = Flatten@StatePosition[system, crit2], matelems, op1},*)
        (*matelems = Tuples[{pos1, pos2}];*)
        (*op1 = ReplaceAllAt[op, rep, matelems];*)
        (*matelems = Reverse /@ DeleteCases[matelems, {i_, i_}];*)
        (*ReplaceAllAt[op1, rep /. (x_ -> y_) :> (x -> EC[y]), matelems]*)
    (*];*)

(*Display*)

TensorForm[x_?OperatorQ] := MatrixForm[x];
TensorForm[x_?SphericalTensorQ] := MatrixForm[x];
TensorForm[x_?TensorDecompositionQ] := MatrixForm /@ x;
TensorForm[x_?TensorDecompositionMatrixQ] :=
    MatrixForm[Map[MatrixForm, x, {3}], TableDepth -> 2];
TensorForm[x_?(CartesianQ[#, ArrayQ]&)] := MatrixForm /@ x;
TensorForm[x_?(CartesianQ[#, !ArrayQ[#]&]&)] := x;

(*Matrix multiplication: inner and outer products*)

(*Define CenterDot to be matrix multiplication*)

LabelOperators[a_] :=
    a /. x_ /; SphericalTensorQ[x, OperatorQ] || CartesianQ[x, OperatorQ] :> MapOnComponents[Operator, x];

Options[MatrixMultiply] = {ContractOneByOnes -> True, SumRanks -> True};

MatrixMultiply[a_?ScalarQ, b_, OptionsPattern[]] := MapOnComponents[a #&, b];

MatrixMultiply[a_, b_?ScalarQ, opts___Rule] := MapOnComponents[b #&, a];

MatrixMultiply[a_?OneByOneQ, b_?OneByOneQ, OptionsPattern[]] :=
    If[OptionValue@ContractOneByOnes, Component, Identity]@{{MatrixMultiply[Component@a, Component@b]}};

MatrixMultiply[a_?OperatorQ, b_?OperatorQ, OptionsPattern[]] = a.b;

MatrixMultiply[a_?OperatorQ, b_ /; CartesianQ[b, OperatorQ], opts : OptionsPattern[]] :=
    MapOnComponents[MatrixMultiply[a, #, opts]&, b];

MatrixMultiply[a_ /; CartesianQ[a, OperatorQ], b_?OperatorQ, opts : OptionsPattern[]] :=
    MapOnComponents[MatrixMultiply[#, b, opts]&, a];

MatrixMultiply[a_?CovariantQ, b_ /; CartesianQ[b, OperatorQ], opts : OptionsPattern[]] :=
    MapOnComponents[MatrixMultiply[a, #, opts]&, b];

MatrixMultiply[a_ /; CartesianQ[a, OperatorQ], b_?ContravariantQ, opts : OptionsPattern[]] :=
    MapOnComponents[MatrixMultiply[#, b, opts]&, a];

MatrixMultiply[a_?CovariantQ, b_ /; CartesianQ[b, ContravariantQ], opts : OptionsPattern[]] :=
    MapOnComponents[MatrixMultiply[a, #, opts]&, b];

MatrixMultiply[a_ /; CartesianQ[a, CovariantQ], b_?ContravariantQ, opts : OptionsPattern[]] :=
    MapOnComponents[MatrixMultiply[#, b, opts]&, a];

MatrixMultiply[a_ /; CartesianQ[a, CovariantQ], b_ /; CartesianQ[b, ContravariantQ], opts : OptionsPattern[]] :=
    Total@MapThread[MatrixMultiply, {a, b}];

MatrixMultiply[a_?TensorDecompositionQ, b_?TensorDecompositionQ,  OptionsPattern[]] :=
    If[OptionValue@SumRanks, Total, Identity]@MapThread[MatrixMultiply, {a, b}];

MatrixMultiply[a_?TensorDecompositionMatrixQ, b_?TensorDecompositionMatrixQ,  OptionsPattern[]] :=
    If[OptionValue@SumRanks, Total@OperatorComponents@#, #]&@MapThread[MatrixMultiply, {a, b}, 2];

MatrixMultiply[a_, b_, c__, opts : OptionsPattern[]] :=
    MatrixMultiply[MatrixMultiply[a, b, opts], c, opts];

allowedMultiplicationOrder[a_, b_] :=
    (CartesianQ[a] && (CartesianQ[b] || OperatorQ[b])) ||
        (CovariantQ[a] && (ContravariantQ[b] || OperatorQ[b])) ||
        (OperatorQ[a] && (CartesianQ[b] || ContravariantQ[b] || OperatorQ[b])) ||
        (ContravariantQ[a] && CovariantQ[b]) ||
        Which[
            (CovariantQ[a] && CovariantQ[b]) || (ContravariantQ[a] && ContravariantQ[b]),
            Message[MatrixMultiply::"coco"],
            (OperatorQ[a] && CovariantQ[b]) || (ContravariantQ[a] && OperatorQ[b]),
            Message[MatrixMultiply::"opco"],
            (CartesianQ[a] && SphericalTensorQ[b]) || (SphericalTensorQ[a] && CartesianQ[b]),
            Message[MatrixMultiply::"sphca"]
        ];

MatrixMultiply[a_, b_, opts : OptionsPattern[]] :=
    (Inner[MatrixMultiply, LabelOperators@a, LabelOperators@b, Plus] /. Operator -> Sequence //
        If[SphericalTensorQ[a] && SphericalTensorQ[b], Replace[#, x_?OneByOneQ :> Component@x], #]&
    ) /; allowedMultiplicationOrder[a, b];

CenterDot = MatrixMultiply;

(*System`CenterDot = CenterDot;*)

TensorNorm[a_ /; SphericalTensorQ[a] || CartesianQ[a]] :=
    Sqrt[ToCovariant[a]\[CenterDot]ToContravariant[a]];

TensorNorm[M_?TensorDecompositionQ] := TensorNorm /@ M;

(*Conjugation*)

(*"Explicit" complex conjugate, assuming all variables are real.*)

ExplicitConjugate[expr_] := expr /. Complex[a_, b_] -> Complex[a, -b];

EC = ExplicitConjugate;

SuperStar = ExplicitConjugate;

ExplicitConjugateTranspose[a_ /; SphericalTensorQ[a] || OperatorQ[a]] :=
    Transpose@MapOnComponents[ExplicitConjugateTranspose, a];

ExplicitConjugateTranspose[a_ /; CartesianQ[a] || TensorDecompositionQ[a] || TensorDecompositionMatrixQ[a]] :=
    MapOnComponents[ExplicitConjugateTranspose, a];

ExplicitConjugateTranspose[a_?ScalarQ] := EC[a];

ECT = ExplicitConjugateTranspose;

GeneralizedConjugateTranspose[a_ /; SphericalTensorQ[a] || OperatorQ[a]] :=
    Transpose@MapOnComponents[GeneralizedConjugateTranspose, a];

GeneralizedConjugateTranspose[a_ /; CartesianQ[a] || TensorDecompositionQ[a] || TensorDecompositionMatrixQ[a]] :=
    MapOnComponents[GeneralizedConjugateTranspose, a];

GeneralizedConjugateTranspose[a_?ScalarQ] := Conjugate[a];

GCT = GeneralizedConjugateTranspose;

SuperDagger = GeneralizedConjugateTranspose;

ToCovariant[a_?CovariantQ] := a;

ToCovariant[a_?ContravariantQ] := Dual@a;

ToCovariant[a_?TensorDecompositionMatrixQ] := MapOnComponents[ToCovariant, a];

ToCovariant[a_?TensorDecompositionQ] := ToCovariant /@ a;

ToContravariant[a_?ContravariantQ] := a;

ToContravariant[a_?CovariantQ] := Dual@a;

ToContravariant[a_?TensorDecompositionMatrixQ] := MapOnComponents[ToContravariant, a];

ToContravariant[a_?TensorDecompositionQ] := ToContravariant /@ a;

(*Dual*)

(*Switch representation from covariant to contravariant, or vice versa. [Var88, Sec. 3.1.2, Eq. (10)]*)
(*	Subscript[A, q]=(-1)^q A^-q,	A^q=(-1)^q Subscript[A, -q]*)
(*where A is any real irreducible tensor, including real vectors. An array with dimension greater than one is assumed to be a tensor operator.*)

DualMatrix[rank_] := Reverse@DiagonalMatrix@Reverse@Table[(-1)^q, {q, -rank, rank}];

Dual[a_?OneByOneQ] := a;

Dual[a_?TensorDecompositionMatrixQ] := MapOnComponents[Dual, a];

Dual[a_?TensorDecompositionQ] := Dual /@ a;

Dual[a_?ContravariantQ] := Transpose[DualMatrix[Rank[a]]\[CenterDot]a];

Dual[a_?CovariantQ] := Transpose[a\[CenterDot]DualMatrix[Rank[a]]];

Dual[a_?CartesianQ] := a;

(*Complex conjugate of an irreducible tensor [Var88, Sec. 1.2.1, Eq. (8)]*)
(*	Subscript[( *)
(*\!\(\*SuperscriptBox["A", "*"]\)), q]=*)
(*\!\(\*SuperscriptBox[*)
(*RowBox[{"(", *)
(*SuperscriptBox["A", "q"], ")"}], "*"]\)=(-1)^q *)
(*\!\(\*SuperscriptBox[*)
(*RowBox[{"(", *)
(*SubscriptBox["A", *)
(*RowBox[{"-", "q"}]], ")"}], "*"]\)*)
(*For a vector, this is equivalent to *)
(*	ToSpherical[Conjugate[ToCartesian[A]]]*)

TensorConjugate[a_?SphericalTensorQ] := Transpose@Conjugate@Dual@a;

TensorConjugate[a_?CartesianQ] := Conjugate[a];

(*Spherical to Cartesian conversion*)

(*Any vector A can be expanded in terms of basis vectors:*)
(*	A=\[Sum]A^\[Mu] Subscript[e, \[Mu]]=\[Sum]Subscript[A, \[Mu]] e^\[Mu]*)
(*Subscript[A, \[Mu]] are the covariant components*)
(*	Subscript[A, \[Mu]]=A\[CenterDot]Subscript[e, \[Mu]],*)
(*and A^\[Mu] are the contravariant components*)
(*	A^\[Mu]=A\[CenterDot]e^\[Mu].*)

(*Convert the spherical components of a vector [Var88, Sec. 1.2.1, Eq. (11)] or a set of spherical basis vectors [Var88, Sec. 1.1.5, Eq. (30)] to cartesian representation. The relations are the same in either case.*)
(* The behavior seems pretty arbitrary for kappa > 1. Maybe should set it up so that the components of the tensor in the "Cartesian" basis are all real *)

ToCartesianMatrix[rank_] :=
    Flatten[{
        Table[
            {Table[Switch[j, i, -1, -i, 1, _, 0], {j, rank, -rank, -1}],
                Table[Switch[j, i, -I, -i, -I, _, 0], {j, rank, -rank, -1}]},
            {i, rank, 1 / 2, -1}] / Sqrt[2],
        {If[IntegerQ[rank], {Table[KroneckerDelta[j], {j, rank, -rank, -1}]}, {}]}
    }, 2];

ToCartesian[a_?ContravariantQ] := Components[ToCartesianMatrix@Rank@a\[CenterDot]a];

ToCartesian[a_?CovariantQ] := Components[a\[CenterDot]ECT@ToCartesianMatrix@Rank@a];

ToCartesian[a_?CartesianQ] := a;

cartesianVectorTimesMatrix[vec_?(CartesianQ[#, (OperatorQ[#] || SphericalTensorQ[#]&)]&), mat_?OperatorQ] := Transpose[Transpose[vec, {3, 1, 2}].mat, {2, 3, 1}];

cartesianVectorTimesMatrix[vec_?CartesianQ, mat_?OperatorQ] := vec.mat;

cartesianVectorTimesMatrix[mat_?OperatorQ, vec_?CartesianQ] := mat.vec;

ToContravariant[a_?CartesianQ] := Transpose@{cartesianVectorTimesMatrix[ECT@ToCartesianMatrix@Rank@a, a]};

ToCovariant[a_?CartesianQ] := {cartesianVectorTimesMatrix[a, ToCartesianMatrix@Rank@a]};

ToCartesian[sys : {_AtomicState..}, op_?OperatorQ] := Module[
    {mat = BlockDiagonal[ToCartesianMatrix /@ F /@ RemoveZeemanSublevels[sys]]},
    mat\[CenterDot]op\[CenterDot]ECT[mat]
];

(*Tensor product*)

(*Irreducible tensor product Subscript[\!\(TraditionalForm\`{\**)
(*SubscriptBox[*)
(*StyleBox["A",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], *)
(*SubscriptBox["\[Kappa]", "1"]]\[CircleTimes]\**)
(*SubscriptBox[*)
(*StyleBox["B",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], *)
(*SubscriptBox["\[Kappa]", "2"]]}\), \[Kappa]] [Var88, Sec. 3.1.7, Eqs. (20,21)]*)
(*The tensor product of rank 0 is related to the scalar product by [Var88, Sec. 3.1.8, Eq. (35)]*)
(*	Subscript[A, \[Kappa]]\[CenterDot]Subscript[B, \[Kappa]]=(-1)^\[Kappa] Sqrt[2\[Kappa]+1] Subscript[{Subscript[A, \[Kappa]]\[CircleTimes]Subscript[B, \[Kappa]]}, 00]*)

SphericalTensorProduct[a_?SphericalTensorQ, b_?SphericalTensorQ, {k_}] :=
    {Reverse@Table[SphericalTensorProduct[a, b, {k, q}], {q, -k, k}]};

SphericalTensorProduct[a_?CovariantQ, b_?CovariantQ, {k_, q_}] :=
    Sum[
        Sum[
            CG[{Rank[a], q1}, {Rank[b], q2}, {k, q}] Component[a, q1] Component[b, q2],
            {q2, -Rank[b], Rank[b]}
        ],
        {q1, -Rank[a], Rank[a]}
    ];

End[];

EndPackage[];
